#include <stdlib.h>
#include "mpi.h"

#define SERVER     (n_procs - 1)
#define ROOT       0
#define REQ_POINTS 1000
#define REQUEST    0
#define REPLY      1

main(int argc, char **argv) {
  MPI_Group world_group, worker_group;
  MPI_Comm workers_comm;
  MPI_Status status;
  int my_rank, n_procs;
  double start, finish, x, y;
  int in, out, total_in, total_out, total_points, i, rands[REQ_POINTS], req_points, ranks[1];

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &n_procs);
  if (n_procs == 1) {
    MPI_Finalize();
    if (my_rank == 0)
      printf("Execute o programa com pelo menos 2 processos!\n");
    exit(0);
  }

  // define novo comunicador para os clientes
  MPI_Comm_group(MPI_COMM_WORLD, &world_group);
  ranks[0] = SERVER;
  MPI_Group_excl(world_group, 1, ranks, &worker_group);
  MPI_Comm_create(MPI_COMM_WORLD, worker_group, &workers_comm);
  MPI_Group_free(&worker_group);
  MPI_Group_free(&world_group);

  // obtm o nmero de pontos e propaga a todos
  if (my_rank == ROOT) {
    printf("Indique o total de pontos: ");
    scanf("%d", &total_points);
  }
  MPI_Bcast(&total_points, 1, MPI_INT, ROOT, MPI_COMM_WORLD);

  // inicia contagem do tempo
  MPI_Barrier(MPI_COMM_WORLD);
  start = MPI_Wtime();
  // calcula o valor aproximado de PI
  if (my_rank == SERVER) {  // servidor
    do {
      MPI_Recv(&req_points, 1, MPI_INT, MPI_ANY_SOURCE, REQUEST, MPI_COMM_WORLD, &status);
      if (req_points) {
	for (i = 0; i < req_points; i++)
          rands[i] = random();
	MPI_Send(rands, req_points, MPI_INT, status.MPI_SOURCE, REPLY, MPI_COMM_WORLD);
      }
    } while (req_points);
  } else {  // cliente
    in = out = 0;
    do {
      req_points = REQ_POINTS;
      MPI_Send(&req_points, 1, MPI_INT, SERVER, REQUEST, MPI_COMM_WORLD);
      MPI_Recv(rands, req_points, MPI_INT, SERVER, REPLY, MPI_COMM_WORLD, &status);
      for (i = 0; i < req_points; i += 2) {
        x = (((double) rands[i]) / RAND_MAX) * 2 - 1;
        y = (((double) rands[i+1]) / RAND_MAX) * 2 - 1;
        (x * x + y * y < 1.0) ? in++ : out++;
      }
      MPI_Allreduce(&in, &total_in, 1, MPI_INT, MPI_SUM, workers_comm);
      MPI_Allreduce(&out, &total_out, 1, MPI_INT, MPI_SUM, workers_comm);
      req_points = (total_in + total_out < total_points);
      if (req_points == 0 && my_rank == ROOT)
	MPI_Send(&req_points, 1, MPI_INT, SERVER, REQUEST, MPI_COMM_WORLD);
    } while (req_points);
    MPI_Comm_free(&workers_comm);
  }
  // termina contagem do tempo e escreve resultado
  MPI_Barrier(MPI_COMM_WORLD);
  finish = MPI_Wtime();
  if (my_rank == ROOT) {
    printf("PI = %.20f\n", (4.0 * total_in) / (total_in + total_out));
    printf("Tempo de Execuo = %f segundos\n", finish - start);
  }
  MPI_Finalize();
}
